package org.jdom;

import java.io.Serializable;

import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

/**
 * <code>Text</code> hold character data within the content of an 
 * <code>Element</code>.
 * 
 * @see org.jdom.Element org.jdom.Element
 * @author Todd Trimmer
 * @version 1.0
 */
public class Text implements Serializable, Cloneable, NodeX {
	
	/** Textual data */
	protected String data;
	
	/** 
	 * Parent <code>Element</code>.
	 *  
	 * @see org.jdom.Element org.jdom.Element
	 **/
	protected Element parent;

	/**
	 * <p>
	 * Default ctor.
	 * </p>
	 */
    public Text() {
		data = "";
    }
	
	/**
	 * <p>
	 * String injection ctor
	 * </p>
	 * 
	 * @param string<code>String</code> contents of Text node.
	 */
	public Text(String string) {
		data = "" + string; // Force NPE
	}
	
	/**
	 * <p>
	 * Textual contents as a <code>String</code>.
	 * </p>
	 * 
	 * @return text as <code>String</code>.
	 */
	public String toString() {
	    return data;	
	}
	
	/**
	 * <p>
	 * Make a copy of this Text node.
	 * </p>
	 * 
	 * @return <code>Text</code> copy of this one.
	 */
	public Object clone() {
		Text text = null;
		
	    try {
            text = (Text)super.clone();
        }
		catch(CloneNotSupportedException ce) {
            // Can't happen
        }
		
		text.parent = null;
		
		return text;
	}
	
	/**
	 * <p>
	 * Equivalence test.
	 * </p>
	 * 
	 * @param ob <code>Object</code> to check equality against.
	 * @return <code>boolean</code> indicating equality/non-equality.
	 */
	public final boolean equals(Object ob) {
        return (this == ob);
    }
	
	/**
	 * <p>
	 * Set the character data.
	 * </p>
	 * 
	 * @param string <code>String</code> to set data as.
	 * @return <code>Text</code> reference to self.
	 */
	public Text setText(String string) {
	    //if (string != null)
			data = "" + string;
			
		return this;
	}
	
	/**
	 * <p>
	 * Get character data as <code>String</code>.
	 * </p>
	 * 
	 * @return Character data as <code>String</code>.
	 */
	public String getText() {
	    return data;	
	}
	
	/**
	 * <p>
	 * Get character data as string, with whitespace trimmed off and internal
	 * spaces normalized.
	 * </p>
	 * 
	 * @return <code>String</code> of trimmed text.
	 */
	public String getTextTrim() {
        String text = getText();

        StringBuffer textContent = new StringBuffer();
        StringTokenizer tokenizer = new StringTokenizer(text);
        while (tokenizer.hasMoreTokens()) {
            String str = tokenizer.nextToken();
            textContent.append(str);
            if (tokenizer.hasMoreTokens()) {
                textContent.append(" ");  // separator
            }
        }

        return textContent.toString();
    }
	
	/**
	 * <p>
	 * Concatenate contents of two Text nodes.
	 * </p>
	 * 
	 * @param text <code>Text</code> containing character data to concatenate
	 * to this one.
	 * 
	 * @return <code>Text</code> reference to self.
	 */
	public Text concat(Text text) {
	    data += text.toString();
		
		return this;	
	}
	
	/**
	 * <p>
	 * Concatenate a string to character data already in Text node.
	 * </p>
	 * 
	 * @param string <code>String</code> of character data to concatenate.
	 * @return <code>Text</code> reference to self.
	 */
	public Text concat(String string) {
        data += string;
			
		return this;	
	}
	
	/**
	 * <p>
	 * Get Element that is parent of this Text node.
	 * </p>
	 * 
	 * @return <code>Element</code> parent.
	 */
	public Element getParent() {
		return parent;
	}
	
	/**
	 * <p>
	 * Set Element that is parent of this Text node.
	 * </p>
	 * 
	 * @param parent New <code>Element</code> parent.
	 * @return <code>Text</code> reference to self.
	 */
	protected Text setParent(Element parent) {
        this.parent = parent;
        return this;
    }
	
	/**
	 * <p>
	 * Detach from current parent.
	 * </p>
	 * 
	 * @return <code>Text</code> reference to self
	 */
	public Text detach() {
        Element p = getParent();
		
        if (p != null) {
            p.removeContent(this);
        }
      
        return this;
    }
	
	/**
	 * <p>
	 * Go up to parent node.
	 * </p>
	 * 
	 * @return parent.
	 */
	public NodeX goUp() {
		return getParent();
	}
	
	/**
	 * <p>
	 * Go down to first child node.
	 * </p>
	 * 
	 * @return first child.
	 */
	public List goDown() {
		return new ArrayList();
	}
	
	/**
	 * <p>
	 * Allow a visitor to peek at this node.
	 * </p>
	 * 
	 * @param visitor <code>NodeVisitor</code> wanting to peek at this node.
	 * @return Arbitrary <code>Object</code> returned from the visitor.
	 * @see org.jdom.NodeVisitor org.jdom.NodeVisitor
	 */
	public Object acceptVisitor(NodeVisitor visitor) {
		return visitor.visitText(this);
	}
	
	/**
	 * <p>
	 * Return a <code>NodeIterator</code> that will traverse through all the
	 * children of this node.
	 * </p>
	 * 
	 * @return <code>NodeIterator</code> to traverse this node's children.
	 * @see org.jdom.NodeIterator org.jdom.NodeIterator
	 */
	public NodeIterator nodeIterator() {
	    return new NodeIterator(this);	
	}
	
	public String debug() {
		return "text " + data.substring(0, data.length() <= 8 ? data.length() : 0);
	}
}